/* Copyright (C) 2000 W.H.Scholten

   Redistribution and use in source and binary forms, with or
   without modification, are permitted provided that the following
   conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.
  
   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
   IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
   FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
   AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h> /* getenv */

/* Debugging */
#define DEB(a) { char *db_env = getenv("DEBUG"); int db_l = 0; if (db_env != NULL) db_l = atol(db_env); if (db_l) { a; } }

int whs_alpha_num_cmp(const char *a, const char *b);

int main(int argc, char**argv) {
	int  i;

	struct my_cmp { char *l; char *r; int check_value; int value; }
	s[ ] = {
		"", "", 1, 0,
		"wipeout", "wipeout2097", 1, -1,
		"wipeout", "wipeoutxl", 1, -1,
		"wipeout2097", "wipeoutxl", 0, -1,
		"wipeout2097x", "wipeoutxlxxxxxx", 0, -1,
		"wipeout2097x", "wipeoutxlxxx2098", 0, -1,
		".", "..", 1, -1,
		"test002", "test02", 1, +1,
		"test02", "test0003", 1, -1,
		"test03", "test0002", 1, +1,
		"testa", "testb", 1, -1,
		"testing", "test1", 0, -1,
		"test1.jpg", "testa.jpg", 0, -1,
		"test1.jpg", "testa0.jpg", 0, -1,
		"test1.jpg", "tEST1.jpg", 1, +1,
		"test1.jpg", "TESTING1.jpg", 1, +1,
		"test1.jpg", "TES1.jpg", 1, +1,
		"test02.jpg", "test2.jpg", 1, +1,
		"test2.jpg", "test2.jpg", 1, 0,
		"test2.0", "test2.0", 1, 0,
		"test2.jpg", "test3.jpg", 1, -1,
		"test2a.jpg", "test3.jpg", 1, -1,
		"test2.jpg", "test10.jpg", 1, -1,
		"test2.jpg", "test10a.jpg", 1, -1,
		"test10.jpg", "test10a.jpg", 1, -1,
		"test10a.jpg", "test10b.jpg", 1, -1,
		"rubik-0.9.1.c", "rubik-0.9.1.1.c", 0, -1,
	};

	fprintf(stderr, "\n Comparisons:\n"
			" ------------\n"
			" a value < 0 means left size is smaller than right side.\n");
	for (i=0; i<sizeof(s)/sizeof(struct my_cmp); i++) {
		int  cmp;

		DEB(fprintf(stderr, "\n-- comparison %2d: %s <=> %s:\n", i, s[i].l, s[i].r));
		cmp = whs_alpha_num_cmp(s[i].l, s[i].r);
		fprintf(stderr, "Comparison %2d: %s <=> %s ? => %d", i, s[i].l, s[i].r, cmp);
		if (s[i].check_value) {
			/* Normalize. */
			if (cmp > 0) cmp = 1; else if (cmp < 0) cmp = -1;

			if (cmp != s[i].value) {
				fprintf(stderr, " [ FAILED: normalized %d, required: %d ]", cmp, s[i].value);
			}
			else {
				fprintf(stderr, " [ OK ]");
			}
		}
		else fprintf(stderr, " [ NOT TESTED ]");

		fprintf(stderr, "\n");
	}


	return 0;
}


#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) < (b) ? (b) : (a))

/* alpha - numeric sorting:
   It works as follows: get the longest alpha part (=anything not containing
   0 .. 9) from both strings, compare these. If equal, compare the numerical
   parts that follow e.g. 2 < 10, 0002<10, 02 < 0002, by stripping leading
   zeros, checking string lengths, and finally do a lexicographical comparison
   (strcmp) if these lengths are equal.

   Version 1.2
 */

//#define NUMBERS_LARGER_THAN_ALPHA

int
whs_alpha_num_cmp(a, b)
	const char *a, *b;
{
	int la, lb;
	int ca, cb;
	int na, nb;

	int new_ca, new_cb, new_na, new_nb;

	for (;;) {
		int cmp;

		la = strlen(a);
		lb = strlen(b);

		ca = strcspn(a, "0123456789");
		cb = strcspn(b, "0123456789");

		DEB(fprintf(stderr, "-  la=%d, lb=%d, ca=%d, cb=%d\n", la, lb, ca, cb));

		if (ca >= la || cb >= lb) {
			/* At most one of the strings has digits: */
#ifndef NUMBERS_LARGER_THAN_ALPHA
			DEB(fprintf(stderr, "(1)\n"));
			return strcmp(a, b);
#else
			/* IS THIS OK?: cmp = strncmp(a, b, min(ca, cb)); TO BE SURE: */
			cmp = strncmp(a, b, max(1, min(ca, cb)));
			if (cmp) { /* alpha parts of the strings are different */
				DEB(fprintf(stderr, "(1) strncmp(a,b, %d)\n", min(ca, cb)));
				return cmp;
			}

			DEB(fprintf(stderr, "(1a) la=%d, ca=%d, lb=%d, cb=%d\n", la, ca, lb, cb));
			/* Equal; then check which of the following cases applies:
			 *   1. neither contains digits.
			 *   2. b contains digits.
			 *   3. a contains digits.
			 */
			if (ca >= la && cb >= lb) return la - lb;
			if (ca >= la) return -1;
			return 1;
#endif
		}

		/* Both strings have digits: */
		if (ca != cb) {
			cmp = strncmp(a, b, min(ca, cb));
			if (cmp) { /* alpha parts of the strings are different */
				DEB(fprintf(stderr, "(2) strncmp(a,b, %d)\n", min(ca, cb)));
				return cmp;
			}

			/* So, the alpha parts are the same up to min(ca, cb).
			 *
			 * If cb > ca the rest of b starts with alpha whereas the rest of
			 * a starts with a digit.
			 */

#ifndef NUMBERS_LARGER_THAN_ALPHA
			DEB(fprintf(stderr, "(2a)\n"));
			return a [ min(ca, cb) ] - b [ min(ca, cb) ];
#else
			/* We say: numbers > alpha, so in this case a > b.
			 */
			DEB(fprintf(stderr, "(2a) ca - cb = %d\n", ca - cb));

			return cb - ca;
#endif
		}


		cmp = strncmp(a, b, ca);
		if (cmp) {
			DEB(fprintf(stderr, "(3) strncmp(a,b, %d)=%d\n", ca, cmp));
			return cmp;
		}

		/* ca=cb & alpha parts are the same;
		 * Now we have a look at the numbers and compare them. */
		na = strspn(a+ca, "0123456789");
		nb = strspn(b+cb, "0123456789");

		/* Remove leading 0's */
		new_na = na;
		new_ca = ca;
		while (a [ new_ca ] == '0') { new_na --; new_ca ++; }

		new_nb = nb;
		new_cb = cb;
		while (b [ new_cb ] == '0') { new_nb --; new_cb ++; }

		if (new_na != new_nb) {
			DEB(fprintf(stderr, "(4) new_na - new_nb = %d\n", new_na-new_nb));
			return new_na - new_nb;
		}

		DEB(fprintf(stderr, "-  a+new_ca=%s, b+new_cb=%s, complen=%d\n", (a+new_ca), (b+new_cb), new_na));

		cmp = strncmp( (a+new_ca), (b+new_cb), new_na);
		if (cmp) {
			DEB(fprintf(stderr, "(5) strncmp(a+new_ca,b+new_cb, new_na)\n"));
			return cmp;
		}
		if (na != nb) {
			DEB(fprintf(stderr, "(6) na != nb (one has more leading zero's)\n"));
			return na - nb;
		}

		DEB(fprintf(stderr, "-  continue: a +=%d, b +=%d anew=%s, b_new=%s\n", ca+na, cb+nb, (a+ca+na), (b+cb+nb)));

		a = (a+ca+na);
		b = (b+cb+nb);
	}
}
